#define INCL_DOSFILEMGR          /* File Manager values */
#define INCL_DOSERRORS           /* DOS Error values    */
#include <os2.h>
#include <gl.h>
#include <glu.h>
#include <aux.h>
#include <ithread.hpp>

#include "glwindow.hpp"

#define FOVY            20.0
#define MINVIEWDISTANCE 160.0
#define MAXVIEWDISTANCE 240.0

// GLpaint destructor handler
Boolean GLPaintDestroy::dispatchHandlerEvent( IEvent &event )
{
   Boolean stopDispatching = false;
   switch( event.eventId() )
   {
      case WM_DESTROY:
         stopDispatching = destroy( event );
         break;
   }
   return( stopDispatching );
}

Boolean GLPaintDestroy::destroy( IEvent &event )
{
   // stop the timer
   paintHandler->timer.stop();
   return false;
}

// constructors and destructors
GLPaintHandler::GLPaintHandler( IWindow &_window )
: window( _window ),  paintDestroy( this ), rotationVal( 0.0 ), bitmap( 0 )
{
   // activate the destructor handler
   paintDestroy.handleEventsFor( &_window );

   // set timer
   timer
      .setInterval( 100 )
      .start( new ITimerMemberFn0<GLPaintHandler>( *this, GLPaintHandler::timerFn ));

   // read in our bitmaps
}

GLPaintHandler::~GLPaintHandler()
{
   HAB hab = IThread::current().anchorBlock();

   // unbind the current context
   pglMakeCurrent( hab, NULL, NULL );

   // destroy the context we created
   pglDestroyContext( hab, hgc );

   if( bitmap )
      delete [] bitmap;
}

void GLPaintHandler::initOpenGL()
{
   int attribList[] = { PGL_RGBA, PGL_RED_SIZE, 2, PGL_GREEN_SIZE, 2,
      PGL_BLUE_SIZE, 2, PGL_ALPHA_SIZE, 1, PGL_DOUBLEBUFFER, 0 };

   // get a handle to anchor block
   HAB hab =  IThread::current().anchorBlock();

   // create a visual config
   PVISUALCONFIG pVisualConfig = pglChooseConfig( hab, attribList );

   // create a direct OpenGL context
   hgc = pglCreateContext( hab, pVisualConfig, NULL, true );

   // and make the context we got the current one
   Boolean val = pglMakeCurrent( hab, hgc, window.handle() );

   // enable depth buffer
   glEnable( GL_DEPTH_TEST );

   glClearColor( 0.0, 0.0, 0.0, 1.0 );
   glDrawBuffer( GL_FRONT_AND_BACK );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT  );
   glDrawBuffer( GL_BACK );

   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

   glDepthFunc( GL_LEQUAL );

   // enable aplha test for transparency
   glAlphaFunc( GL_GREATER, 0.5 );
   glEnable( GL_ALPHA_TEST );

   // free the visual config created above
   delete [] pVisualConfig;
}

void GLPaintHandler::initTextureMapping()
{
   // read in
   bitmap = readBitmap( "opengl.bmp", 256, 128 );

   // specify texture parameters
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
   glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
   glEnable( GL_TEXTURE_2D );
}

void GLPaintHandler::initLighting()
{
   glEnable( GL_LIGHTING );
   // setup light 0
   glEnable( GL_LIGHT0 );
   glEnable( GL_COLOR_MATERIAL );

   // modify ambient light
   GLfloat amb_light[] = { 0.1, 0.1, 0.1, 1.0 };
   glLightModelfv( GL_LIGHT_MODEL_AMBIENT, amb_light );
}

char * GLPaintHandler::readBitmap( char *fileName, int width, int height )
{
   HFILE  hfFile;
   ULONG  ulAction;
   ULONG  ulBytesRead;
   ULONG  ulWrote;
   ULONG  ulLocal;

   // open the file
   DosOpen( fileName, &hfFile, &ulAction, 0L, FILE_ARCHIVED | FILE_NORMAL,
                 OPEN_ACTION_FAIL_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
                 OPEN_FLAGS_NOINHERIT | OPEN_SHARE_DENYNONE  |
                 OPEN_ACCESS_READONLY, 0L);

   // get the length of the file
   ULONG filePtr, filePtr2;
   DosSetFilePtr( hfFile, 0, FILE_END, &filePtr );

   // read in the file
   char *buffer = new char[filePtr];
   DosSetFilePtr( hfFile, 0, FILE_BEGIN, &filePtr2 );
   DosRead( hfFile, buffer, filePtr, &filePtr2 );

   DosClose( hfFile );

   char *imageBuf = new char[width*height*3];
   // reset data back to RGBA from BGR
   for( int i=0; i<width*height*3; i+=3 )
   {
      imageBuf[i] = buffer[28+i];
      imageBuf[i+1] = buffer[27+i];
      imageBuf[i+2] = buffer[26+i];
   }
   delete [] buffer;
   return( imageBuf );
}

// openGL settings
void GLPaintHandler::setWindowSize( ISize size )
{
   // setup the 3D perspective window
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity();
   gluPerspective( FOVY, (float)window.size().width() / (float)window.size().height(),
                  MINVIEWDISTANCE, MAXVIEWDISTANCE );
   glMatrixMode( GL_MODELVIEW );
   // make a viewport the acual size of the window
   glViewport( 0, 0, size.width(), size.height() );
}

Boolean GLPaintHandler::paintWindow( IPaintEvent &event )
{
   // clear the buffer
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // no rotations
   glLoadIdentity();

   // place a headlight
   GLfloat light_pos[] = { 0, 0, 0, 1 };
   glLightfv( GL_LIGHT0, GL_POSITION, light_pos );

   // move the image back from the camera
   glTranslatef( 0.0, 0.0, -200.0 );
   // and rotate for the viewing position
   glRotatef( rotationVal, 0.0, 1.0, 0.0 );

   // draw the bitmap
   glTexImage2D( GL_TEXTURE_2D, 0, 4, 256, 128, 0, GL_RGB, GL_UNSIGNED_BYTE, (void *)bitmap );
   glBegin( GL_QUADS );
      glNormal3f( 0.0, 0.0, 1.0 );
      glTexCoord2f( 0.0, 0.0 );
      glVertex3f( -32.0, -16.0, 0.0 );
      glTexCoord2f( 1.0, 0.0 );
      glVertex3f( 32.0, -16.0, 0.0 );
      glTexCoord2f( 1.0, 1.0 );
      glVertex3f( 32.0, 16.0, 0.0 );
      glTexCoord2f( 0.0, 1.0 );
      glVertex3f( -32.0, 16.0, 0.0 );
   glEnd();

   // swap the back draw buffer to the front
   pglSwapBuffers( IThread::current().anchorBlock(), window.handle() );

   // return true to indicate no additional processing needed
   return( true );
}

//private objects
void GLPaintHandler::timerFn()
{
   // recalculate positions
   rotationVal += 2.0;
   if( rotationVal > 360.0 )
      rotationVal -= 360.0;

   window.refresh();
}

// window resize handler
Boolean GLResizeHandler::windowResize(IResizeEvent &event )
{
   // tell the paint handler the new size
   window->glPaintHandler().setWindowSize( event.newSize() );
   return false;
}

// constructor
GLWindow::GLWindow()
: IFrameWindow( "OpenGL Demonstration 7", 0x1000, IFrameWindow::defaultStyle()
         | IFrameWindow::noMoveWithOwner | IWindow::synchPaint ),
   canvas( 0x8008, this, this ),
   paintHandler( canvas ),
   resizeHandler( this )
{
   // setup
   setClient( &canvas );
   setDestroyOnClose( true );

   // draw on screen
   sizeTo( ISize( 400, 400 ));
   setFocus();
   show();

   // initialize openGL
   paintHandler.initOpenGL();

   // initialize lighting
   paintHandler.initLighting();

   // initialize openGL
   paintHandler.initTextureMapping();

   // set the size of the 3D window
   paintHandler.setWindowSize( canvas.size() );

   // attach the handlers
   paintHandler.handleEventsFor( &canvas );
   resizeHandler.handleEventsFor( &canvas );
}

// destructor
GLWindow::~GLWindow()
{
   // detach the handlers
   paintHandler.stopHandlingEventsFor( &canvas );
   resizeHandler.stopHandlingEventsFor( &canvas );
}
